from __future__ import annotations

from pyarrow import Schema, RecordBatch  # type: ignore

class ArrowToPostgresBinaryEncoder:
    def __init__(self, __schema: Schema) -> None: ...
    def write_header(self) -> bytes: ...
    def write_batch(self, __batch: RecordBatch) -> bytes: ...
    def finish(self) -> bytes: ...
